@@ -1,4 +1,7 @@
+require_relative "../../../spec_helper"
 platform_is :windows do
+  verbose, $VERBOSE = $VERBOSE, nil
+
   require_relative '../fixtures/classes'
 
   describe "WIN32OLE.connect" do
@@ -12,4 +15,6 @@ platform_is :windows do
     end
 
   end
+ensure
+  $VERBOSE = verbose
 end
