@@ -1,4 +1,7 @@
+require_relative "../../../spec_helper"
 platform_is :windows do
+  verbose, $VERBOSE = $VERBOSE, nil
+
   require 'win32ole'
 
   describe "WIN32OLE_METHOD.new" do
@@ -30,4 +33,6 @@ platform_is :windows do
 
   end
 
+ensure
+  $VERBOSE = verbose
 end
